/* 
	Must be universally compatible 
	WebKit, IE, Firefox, JSContext
	
	Version 1.0 Initial Release
	- Basic wiki style syntex input.
	
	Version 1.1 Sept 2, 2010
	- Preliminary Richtext support
	- Basic Widget Tool Integration.
	
*/

var DWiki = {};

// !DWikiWidget
/*
	Represents a widget base class for use with the DWikiEditor,
	
	BoldWidget, ItalicWidget, ChatWidget, MapWidget, FileWidget, LinkWidget...
	
	Internally this abstract base class implements -bold, -italic, -underline...
*/
fClass.NSObject.extend('DWikiWidget', { ivars: {

id:"bold",
tooltip:"Make text bold.",
icon:null,
char:"B", //Temporary for testing.
editor:null,
showsButton:true, //overide for widget subclasses that have no display button.

/*
	Subclasses can overide this to return true then the caret
	should exist out of the widget, breaking out the trailing text into
	plain text
	
	For example
	<element>Some Content[break-hear] Here</element>
	if true, Here gets appended after a new line
*/
newLineEscapesRange:false

}, methods : {

initWithId:function(widgetId) {
	if(this.init()) {
		this.id = widgetId;
		this.tooltip = "Make text " + widgetId;
		this.icon = widgetId + "_wiki_widget.png";
		this.char = widgetId.charAt(0).toUpperCase();
	}
	return this;
},

/*
	Subclasses should overide this to callout or display any user interface element to
	display if the widget requires further action from the user. Subclasses should call,
	editor.applyWidgetAction(this) to fire the -apply method.
*/
doActionForEditor:function(editor, range) {
	//by defualt we have no interface, so we just apply the action.
	this.editor = editor;
	editor.applyWidgetAction(this, range);
},


/*
	Subclasses should overide the apply function to apply the widget action to the document.
	The attributes are,
	
	- range, A object with key location and key length represent the positioning in the document to
		apply the widget. a zero length indicates a collapsed caret location.
	- line, The line withing the document to apply positioning.
	- text, The document text
	
	The function should return an array with the following objects.
	1. The new text with changes.
	2. A new range (or the existing range).
*/	
apply:function(range, line, text) {
	if(this.id == 'bold') {
		if(this.editor.richTextEnabled) {
			return this.rangeWrap(text, range, "<b>", "</b>");
		}
		return this.rangeWrap(text, range, "*", "*");
	} else if(this.id == 'italic') {
		if(this.editor.richTextEnabled) {
			return this.rangeWrap(text, range, "<i>", "</i>");
		}
		return this.rangeWrap(text, range, "_", "_");
	} else if(this.id == 'underline') {
		if(this.editor.richTextEnabled) {
			return this.rangeWrap(text, range, "<u>", "</u>");
		}
	} else if(this.id == 'strike') {
		if(this.editor.richTextEnabled) {
			return this.rangeWrap(text, range, "<s>", "</s>");
		}
	}
	return [text, range];
},

//RichText Encoding methods, for subclasses

htmlToOefText:function(string) {
	return string;
},

oefTextToHtmlText:function(string) {
	return string;
},

//Utility methods

/*
	Wraps the given text with the before and after characters, returning an array
	containing, 1 = the new text, 2 = the range adjusted to include the before and after insertions.
*/
rangeWrap:function(text, range, before, after) {
	nRange = {};
	nRange.location = range.location;
	nRange.length = range.length;

	text = text.substr(0,range.location)+before+text.substr(range.location);
	range.length += before.length;
	text = text.substr(0,range.location+range.length)+after+text.substr(range.location+range.length);
	
	NSLog("before length:"+before.length+ " after length:"+after.length);
	
	if(!this.editor.richTextEnabled) {
		nRange.length += (after.length+before.length);	
	} else {
		//We wrap directly with elements, so we only need to nudge forward
		nRange.length+=1;
	}
		
	return [text, nRange];
},

insertAt:function(text, location, insertText) {
	return text.substr(0, location) + insertText + text.substr(location);
}

}});

// !DWikiAttachmentWidget
/*
	Provides the chat widget for inserting and replying to documents. This actually is not
	the only way a chat can be inserted, chats can be inserted by replys in the display view
	using html chat widget boxes.
*/
fClass.DWikiWidget.extend('DWikiAttachmentWidget', { ivars: {

attachmentText:null,
editor:null

}, methods: {

doActionForEditor:function(editor, range) {
	this.editor = editor;
	this.range = range;
	var that = this;
	window["applyCallbackForAttachmentWidget"] = function(attachmentText) {
		that.attachmentText = attachmentText;
		that.editor.applyWidgetAction(that, that.range);
	}
	
	//Get caret location.
	
	if(TaskListController) {
		cloc = editor.getCaretLocationInView();
		TaskListController.showWidgetDialogforDNoteAttachmentType_top_left_(this.id, cloc.top, cloc.left);
	}
},

apply:function(range, line, text) {	
	//Auto apply to it's own line, as all attachments are registered by their own line.
	var collapsedIndex = range.location+range.length;
	if((text.charAt(collapsedIndex+1)=='\n'|| text.length == collapsedIndex+1) 
	&& (text.charAt(collapsedIndex-1)=='\n'|| collapsedIndex-1 < 0)) {
		text = this.insertAt(text, collapsedIndex, this.attachmentText);
	} else { //Add to it's own newline
		text = this.insertAt(text, collapsedIndex, "\n"+this.attachmentText+"\n");
	}
	
	return [text, range];
},

//Move these to a subclass?
htmlToOefText:function(string) {
	//We do some jquery parsing here to condense things down rather then trying to run regex,
	//The format of a chat in html is 
	//<div class='comment'><TAG class='cmessage'>'$MESSAGE</TAG><TAG class='csender'>$SENDER</TAG></div>
	//gets converted to ((chat: $MESSAGE -$SENDER)) in attachment syntext form
	
	var stringToEl = $("<div display='none'>"+string+"</div>");
	stringToEl.find("div.comment").each(function(index) {
		var message = $(this).find('.cmessage').text();
		var sender = $(this).find('.csender').text();
			
		/*
			We need to check for newline requirements here. This is a requirements for all line attachments
			so we can eventually move this somewhere else.
		*/
		if($(this).next("br").length==0) {
			$(this).after("<br>");
		}
			
		//Do raw replace, we dont care about newlines for chat messages any more, their automatically put
		//on newlines in each client by design.
		$(this).after("((chat: "+message+" -"+sender+"))").remove();
	});
	return stringToEl.html();
},

oefTextToHtmlText:function(string) {
	//Shortcut mode ^-- Line of text goes here.
	
	//Requires some sort of selection range fix
	//string = string.replace(/^-- (.+)/gm, "((chat: $1 -"+GetMyUsername()+"))");
	
	return string.replace(/\(\(chat\: ([A-Za-z\.\!\@\#\s)]+) -([a-zA-Z0-9-)]+)\)\)/g, "<div class='comment attachment line-based' type='chat'><span class='cmessage'>$1</span><span class='csender'>$2</span></div>");
}

}});

// !DWikiTaskWidget
/*
	Task Widget
*/
fClass.DWikiAttachmentWidget.extend('DWikiTaskWidget', { ivars: {

//Task widget currently doesn't have a toolbar button, and is used primarily for parsing.
showsButton:false,
newLineEscapesRange:true

}, methods: {

htmlToOefText:function(string) {
	var stringToEl = $("<div display='none'>"+string+"</div>");
	stringToEl.find("div.inline-task").each(function(index) {
		var content = $(this).text();
		var state = $(this).hasClass("complete");
			
		/* 
			We need to check for newline requirements here. This is a requirements for all line attachments
			so we can eventually move this somewhere else.
		*/
		if($(this).next("br").length==0) {
			$(this).after("<br>");
		}
		
		$(this).after(((state)?"+ ":"- ")+content).remove();
	});

	return stringToEl.html();
},

oefTextToHtmlText:function(string) {
	return string.replace(/^- (.+)/gm, "<div class='inline-task attachment line-based' type='task'>$1</div>").replace(/^\+ (.+)/gm, "<div class='inline-task attachment line-based complete' type='task'>$1</div>")
},

}});

// !DWikiEditor
/*
	fClass wrapper on the Editor view
	This will eventuall replace the existing DWiki info to use fClass for complete encapsulation
	of content editable area
	
	Notes:
	As of version 1.1 preliminary richtext support is enabled. For rich text support, the
	defualt safari/webkit richtext content editable is disabled, and we use our own custom system.
	
	The new system uses our oef file protocal for wrapping formatting into the document's headers.
	Currently DNote only supports the presets such as
	Bold, Italic, Headers 1-9, Underline, Strike and Basic Colors, Custom formatting tags are not
	supported yet.
	
	For display and parsing of display text, we use the following html element types
	Bold 		-> <b> \u12F00 \
	Italic 		-> <i>
	Strike		-> <s>
	Underline 	-> <u>
*/
fClass.NSObject.extend('DWikiEditor', { ivars: {

//Unique reference for the edit, when multiple editors are one one page/view, this is typically a task guid
ref:null,

//Backing content editible div element used for editing.
element:null,

//Pulled in from the element's placeholder attribute.
placeholder:null,

//Unparsed raw wiki text
wikitext:null,

//A seed key for the sync context. This is set async once key information is received.
seed:null,

//A Callback funtion that gets called when an edit completes. (LEGACY)
changeSelector:null,

//If were in edit mode, or view mode.
editing:false,

widgetContainer:null,

//If richtext entry mode is enabled. This is a more difficult design, as
//no mode switching happens between plain an rich modes, and all text
//is parsed from the backing html.
richTextEnabled:false,

//A dictionary of widgets by widget identifier.
widgets:{},

//Token table used for formatting
tagToTokenTable:{
	"<b>":"\uEF00","</b>":"\uEF01",
	"<i>":"\uEF02","</i>":"\uEF03",
	"<u>":"\uEF04","</u>":"\uEF05",
	"<s>":"\uEF06","</s>":"\uEF07",
}
	
}, methods : {

p_reverseTokenLookup:function() {
	var tokenToTag = {};
	for(var tag in this.tagToTokenTable) {
		tokenToTag[this.tagToTokenTable[tag]] = tag;
	}
	return tokenToTag;
},

/*
	Provides OEF richtext encoding sutable for diff_match_patch diffing for changes.
	Currently this only reconizes basics such as
	<b> <s> <i> <br> ... as used by our plaintext content editable div
*/
oefTextFromHtmlString:function(string) {	
	//NSLog("raw html:"+string);
	
	//Parse out the carets (dks, dke) (as {START} and {END} placeholders) (Always first)
	string = string.replace(/\<\/dks\>/, "").replace(/\<\/dke\>/, "").replace(/\<dks\>/, "{START}").replace(/\<dke\>/, "{END}");
	
	var tag;
	for(tag in this.tagToTokenTable) {
		string = string.replace(new RegExp(tag, 'g'), this.tagToTokenTable[tag]);
	}
	
	//Parse with each widget plugin, this is done first from html
	for(var i in this.widgets) {
		string = this.widgets[i].htmlToOefText(string);
	}
	
	//Breaks are always last becouse some widgets are dependent on looking for breaks for rectifying
	//divs to newlines.
	string = string.replace(/\<br\>/g, "\n").replace(/&nbsp;/g, " ").replace(/\<paste\>/g, "").replace(/\<\/paste\>/, "");
		
	return string;
},

htmlStringFromOefText:function(string) {
	//Caret placements (these get removed on all updates and transmits)
	string = string.replace(/\{START\}/g, "<dks/>").replace(/\{END\}/g, "<dke/>");
	
	//bold+italic+strike+underline (From Wiki Text)
	string = string.replace(/\*([\w\?\.,\/#_`\-\(\) ]+)\*/g, "<b>$1</b>").replace(/_([\w\?\.,\/#_`\-\(\) ]+)_/g, "<i>$1</i>");
	
	//Get the tokenToTagTable
	var tokenTable = this.p_reverseTokenLookup();
	for(var token in tokenTable) {
		string = string.replace(new RegExp(token, 'g'), tokenTable[token]);
	}
	
	//Parse with each widget plugin
	for(var i in this.widgets) {
		string = this.widgets[i].oefTextToHtmlText(string);
	}
	
	//Newlines are always imply a <br> element
	var string = string.replace(/\n/g, "<br>");
	
	return string;
},

//------- Utilities and Parsers [PLAIN-TEXT (DEPRECIATED?)] -----------

htmlTaskFromWikiLine:function(line) {
	var ret = {};
	ret.changed = 0;
	//- incomplete tasks
	if(line.search(/(^|\n)- (.+)($|\n)/) != -1) { //([\w-\(\)#\*\._ ]+)
		ret.line = line.replace(/(^|\n)- (.+)($|\n)/, "<div class='inline-task'>$2</div>");
		ret.changed = 1;
	}
	//+ complete tasks
	if(line.search(/(^|\n)\+ (.+)($|\n)/) != -1) {
		ret.line = line.replace(/(^|\n)\+ (.+)($|\n)/, "<div class='inline-task complete'>$2</div>");
		ret.changed = 1;
	}
	
	if(!ret.changed) {
		ret.line = line;
	}
	return ret;
},

htmlFromWikiString:function(string) {
	if(!string || string.length < 1) {
		//Hack for placeholder string.
		if(this.placeholder) {
			return "<span class='placeholder' style='-webkit-user-modify: read-only'>"+this.placeholder+"</span>";
		} else {
			return "";
		}
	}

	//First handle all block based syntex items (inline tasks, h1-9)
	var changed = 0;
	var lines = string.split('\n');
	for(var i=0;i<lines.length;i++) {
		var line = lines[i].replace(/\s*$/, '');
		changed = 0;
		
		//tasks
		var res = this.htmlTaskFromWikiLine(line)
		if(res.changed == 1) {
			line = res.line;
			changed = 1;
		}

		//h1-6 headers
		if(line.search(/(^|\n)h([1-6])\. (.+)($|\n)/) != -1) {
			line = line.replace(/(^|\n)h([1-6])\. (.+)($|\n)/, "<h$2>$3</h$2>");
			changed = 1;
		}
		
		//Block Comments
		if(line.search(/(^|\n)(\(\(([\w\"\'#\_\-\/\*,\.\[\]\: ]+)\)\))($|\n)/) != -1) {
			line = line.replace(/(^|\n)(\(\(([\w\"\'#\_\-\/\*,\.\[\]\: ]+)\)\))($|\n)/, "<div class='comment'><span class='c-text'>$3</span></div>");
			changed = 1;
		}
	
		//Block Targeted comments
		if(line.search(/(^|\n)\(\(@([\w\-]+) ([\w\"\'#\_\-\/\*,\.\[\]\: ]+)\)\)($|\n)/) != -1) {
			line = line.replace(/(^|\n)\(\(@([\w\-]+) ([\w\"\'#\_\-\/\*,\.\[\]\: ]+)\)\)($|\n)/, "<div class='comment' target='$2'><span class='c-target'>$2</span><span class='c-text'>$3</span></div>");
			changed = 1;
		}
		
		//newlines
		if(changed == 0) {
			line = line + "\n<br>";
		}
		lines[i] = line;
	}
	
	string = lines.join('\n');
	
	//Labels
	string = string.replace(/(^| )#([a-zA-Z0-9]+)($|\n| )/g, "<span class='label' id='$2'>$2</span>");

	
	//Comments
	string = string.replace(/\(\(([\w\"\'#\_\-\/\*,\.\[\]\: ]+)\)\)/g, "<span class='comment'><span class='c-text'>$1</span></span>");
	
	//Targeted comments
	string = string.replace(/\(\(@([\w\-]+) ([\w\"\'#\_\-\/\*,\.\[\]\: ]+)\)\)/g, "<span class='comment' target='$1'><span class='c-target'>$1</span><span class='c-text'>$2</span></span>");
	
	//links
	//string = string.replace(/\[(\))
	
	//bold+italic+strike+underline
	string = string.replace(/\*([\w\?\.,\/#_`\-\(\) ]+)\*/g, "<b>$1</b>").replace(/_([\w\?\.,\/#_`\-\(\) ]+)_/g, "<i>$1</i>");

	//Colors
	string = string.replace(/-([\w\#\(\),\.0-9]+)-([\w\?\.,\/# ]+)-\/([\w\#\(\),\.0-9]+)-/g, "<span style='color:$1'>$2</span>");
	
	return string;
},

wikiFromHtmlString:function(string) {
	//	Does a best approach to method for converting wiki form plain
	string = string.replace(/\<\/dks\>/, "").replace(/\<\/dke\>/, "").replace(/\<dks\>/, "{START}").replace(/\<dke\>/, "{END}");
	string = string.replace(/\<br\>/g, "\n").replace(/&nbsp;/g, " ").replace(/\<paste\>/g, "").replace(/\<\/paste\>/, "");
	return string;
},

//-----------------------------------------

initWithElement:function(element, wikitext, docid, changeSelector) {
	if(this.init()) {
		this.ref = docid;
		this.element = element;
		this.changeSelector = changeSelector;
		this.placeholder = $(element).attr('placeholder');
		
		var that = this;
		$(this.element).focus(function() {
			that.beginEditMode();
		}).blur(function () {
			that.endEditMode();
		}).bind('paste', function(e) {
			that.paste(e);
		});
		
		//Setup rich text tools.
		
		if(this.richTextEnabled) {
			$(this.element).addClass('rich-text-mode')
		} else {
			$(this.element).removeClass('rich-text-mode')
		}
		
		
		this.setWikiText(wikitext);
		
		//TODO: this will be changed for widgets
		$(this.element).bind('inline-task-change', function(e) {
			if(that.changeSelector) {
				that.changeSelector(that.wikitext, that.ref);
			}
		}).keyup(function(event) {
			that.keyUp(event);
		});
		
		/*
			Setup widget interface.
			Currently widgets are simply added to
			an element that is set as self's widgetContainer
		*/
		this.widgetContainer = $(this.element).parent().find('.dwiki-widgets');
		this.widgetContainer.mousedown(function(e) {
			e.preventDefault();
			e.stopPropagation();
		}).mouseup(function(e) {
			e.preventDefault();
			e.stopPropagation();
		});
		
		//Defualt widgets.
		this.addWidget(fClass.DWikiWidget.alloc().initWithId('bold'));
		this.addWidget(fClass.DWikiWidget.alloc().initWithId('italic'));
		
		//Defualts for richtext mode
		if(this.richTextEnabled) {
			//Were always in entry mode
			$(this.element).addClass('entrymode')
			this.addWidget(fClass.DWikiWidget.alloc().initWithId('underline'));
			this.addWidget(fClass.DWikiWidget.alloc().initWithId('strike'));
			this.addWidget(fClass.DWikiTaskWidget.alloc().initWithId('task'));
		} else {
			this.widgetContainer.hide();
		}
		
		//DNote Attachments
		this.addWidget(fClass.DWikiAttachmentWidget.alloc().initWithId('chat'));

	}
	return this;
},

//Returns the mobwrite document key, this is typically fetched from the dejumble server for auth purposes.
syncKey:function() {
	if(this.seed) {
		return 'd'+this.seed.replace(/-/g, '').replace(/\//g, '').replace(/\+/g, '').replace(/\=/g, '')+this.ref.replace(/-/g, '');
	}
	return null;
},

/*
	Return text sutable for a contentEditable div, where \n are converted to <br>
*/
contentEditableWikiText:function(wikiText) {
	var string = wikiText.replace(/\n/g, "<br>");
	
	//Caret placements (these get removed on all updates and transmits)
	string = string.replace(/\{START\}/g, "<dks/>").replace(/\{END\}/g, "<dke/>");
	return string;
},

//No formating, no carets, no oef tokens.
rawText:function() {
	return $(this.element).text();
},

//If the edit area has new text, checks it agains it's placeholders
hasEditedText:function() {
	return (this.rawText() != this.placeholder);
},

/*
	Returns the transmitable, cleaned wiki text
	Note: this works independent of if the editor is in edit mode or display mode.
*/
wikiText:function() {
	if(!this.richTextEnabled) {
		if(this.editing) {	
			return $("<p>"+this.wikiFromHtmlString($(this.element).html())+"</p>").text();
		}
		return this.wikitext;
	} else {
		//Wrapping it then calling -text does addition cleanup
		if(this.editing) {
			return $("<p>"+this.oefTextFromHtmlString($(this.element).html())+"</p>").text();
		} else {
			return this.wikitext;
		}
	}
	return this.wikitext;
},

//This is expected an oefFormatted or DWiki text, it then parses for display, and cleans out any html.
setWikiText:function(wikiText) {
	//Require initialized string.
	if(!wikiText || wikiText.length == 0)
		wikiText = "";

	if(this.richTextEnabled) {
		//Editing and display mode are all the same now, just parse the text into something displayable, then set our html to it.
		$(this.element).html(this.htmlStringFromOefText(wikiText));	
	} else {
		if(/*$(this.element).hasClass('entrymode')*/this.editing) {
			$(this.element).html(this.contentEditableWikiText(wikiText));
		} else {
			$(this.element).html(this.htmlFromWikiString(wikiText));
		}
	}

	//Always store the raw transmitable version
	this.wikitext = $("<p>"+wikiText+"</p>").text();
},

refreshWithTextStorage:function() {
	var offsets = this.getSelectionOffsets();
	this.setWikiText($("<p>"+this.oefTextFromHtmlString($(this.element).html())+"</p>").text());
	if(this.editing) {
		this.setSelectionOffsets(offsets);
	}
},

//---------- Events -------------

beginEditMode:function() {
	this.editing = true;
	
	//Nothing changes, we dont need to update are view, as it stays put.
	if(!this.richTextEnabled) {
		this.setWikiText(this.wikitext);
		$(this.element).addClass('entrymode')
	}
},

//Blur, we write our changes back to cocoa now as a save commit.
//This may also get called rapidly when formatting occurs, or an attachment is added to the document.
endEditMode:function() {
	this.editing = false;
	
	if(this.richTextEnabled) {
		if(this.hasEditedText()) { //placeholder support
			//Just update the backing wikitext storage
			this.wikitext = $("<p>"+this.oefTextFromHtmlString($(this.element).html())+"</p>").text();
			this.setWikiText(this.wikitext);
		} else {
			this.wikitext = "";
		}
	} else {
		$(this.element).removeClass('entrymode');
		if(this.rawText() != this.placeholder) {
			this.setWikiText(this.wikiFromHtmlString($(this.element).html()));
		} else {
			this.wikitext = "";
		}
	}
	
	//Send a callback (will be moved to a delegate system eventually), always the opaque -wikitext property.
	if(this.changeSelector) {
		this.changeSelector(this.wikitext, this.ref);
	}
},

enter:function(event) {
	/*
		Is their no better way?
		Ugly hack to break out of a attachment widget if needed. Cant seem to get the location here, 
		We need to put a pin point as the break, check if were in a widget, the break out overiding defualt. 
	*/
	var range = this.getSelectedRange();
	var shadowElement = this.elementByInsertingString("<bp/>");
	
	//Check if were breaking in a attachment.
	var bp = shadowElement.find('bp');
	var parentAttachment = bp.closest('.attachment');
	if(parentAttachment.length > 0) {
		var widgetName = parentAttachment.attr("type");
		
		window.console.log(widgetName);
		if(this.widgets[widgetName].newLineEscapesRange) {
			
			var bpElement = bp[0];
			var trailingText = "";
			if(bpElement.nextSibling) {
				//Get the next node after the bp placeholder
				trailingText = bpElement.nextSibling.nodeValue;
				
				//Cut it off the end of the node
				bpElement.nextSibling.nodeValue = "";
			}
			parentAttachment.after("<br>"+trailingText);
			bp.remove();
			
			var newText = this.oefTextFromHtmlString(shadowElement.html());
			
			NSLog("newTExt:"+newText);
			
			this.setWikiText(newText);
			range.location += 1;
			range.length = 0;
			this.setSelectedRange(range);
			//Move our caret over by length of trailingText.
			
							
			event.stopPropagation();
			event.preventDefault();
			return;
		}
	}
		
	this.refreshWithTextStorage();
},

keyUp:function(event) {
	if(event.keyCode == '13') {
		this.enter(event);
	}
},

//Utilities

/*
	Gets the current selection range withing the editor's element
*/
getRange:function() {
 	// Don't capture selection outside editable region
	var isOrContainsAnchor = false,
        isOrContainsFocus = false,
        sel = window.getSelection(),
        parentAnchor = sel.anchorNode,
        parentFocus = sel.focusNode,
        range = null;
        
    while(parentAnchor && parentAnchor != document.documentElement) {
        if(parentAnchor == this.element) {
            isOrContainsAnchor = true;
        }
        parentAnchor = parentAnchor.parentNode;
    }
    
    while(parentFocus && parentFocus != document.documentElement) {
        if(parentFocus == this.element) {
            isOrContainsFocus = true;
        }
        parentFocus = parentFocus.parentNode;
    }
    
    if(!isOrContainsAnchor || !isOrContainsFocus) {
        return;
    }
    
    var selection = window.getSelection();
    
    // Get range (standards)
    if(selection.getRangeAt !== undefined) {
        range = selection.getRangeAt(0);

    // Get range (Safari 2)
    } else if(
        document.createRange &&
        selection.anchorNode &&
        selection.anchorOffset &&
        selection.focusNode &&
        selection.focusOffset
    ) {
        range = document.createRange();
        range.setStart(selection.anchorNode, selection.anchorOffset);
        range.setEnd(selection.focusNode, selection.focusOffset);
    } else {
        // Failure here, not handled by the rest of the script.
        // Probably IE or some older browser
    }
    
    return range;
},

/*
	Returns a object containing the following key values.
	location: the index of the selection start within the document text.
	length: the length of the selection, or zero if single caret position.
*/
getSelectedRange:function() {
	var offsets = this.getSelectionOffsets();
	var range = {location:0,length:0}
	range.location = offsets.selectionStart;
	range.length = (offsets.selectionEnd - offsets.selectionStart);
	return range;
},

setSelectedRange:function(range) {
	var offsets = {selectionStart:range.location,selectionEnd:range.location+range.length};
	this.setSelectionOffsets(offsets);
},

/*
	Returns an x,y or top, left position for the caret within the editor view
	This does some fancy backend work to insert a hidden element, then uses jquery
	to get the element's display corredents, then removes the hidden element.
*/
getCaretLocationInView:function() {
	var sel = window.getSelection();
	var range = sel.getRangeAt(0);
	range.insertNode($("<clsv/>")[0]);
	//range.setStartAfter(newTextNode);
	//sel.removeAllRanges();
	//sel.addRange(range);
	
	
	//TaskCreateController.showProjectCreateAtTop_left_($(document).height()-$(this).offset().top-($(this).height()/2)-2, $(this).offset().left+36-4)
	
	loc = {};
	loc.top = $(document).height()-$('clsv').offset().top;
	loc.left = $('clsv').offset().left;
	$('clsv').remove();
	
	return loc;

},

/*
	Returns the selection offsets using and advanced method of inserting karets, then parsing the 
	range from the carets.
*/
getSelectionOffsets:function() {
  var sel = {};
  
  var range = this.getRange();
  if(!range) {
  	return null;
  }
  
  var cstart = document.createElement('dks');
  range.insertNode(cstart);
  
  if(!range.collapsed) {
  	var cend = document.createElement('dke');
	range.collapse();
	range.insertNode(cend);
  }
  
  var text = this.wikiText();
  sel.selectionStart = text.indexOf('{START}', 0);
  text = text.replace("{START}", "");
  sel.selectionEnd = text.indexOf('{END}', sel.selectionStart); 
  
  //Remove selection refs
  $('dks').remove();
  $('dke').remove();
  
 // NSLog('Selection Start:'+sel.selectionStart + ' End:' + sel.selectionEnd);
  return sel;
},

setSelectionOffsets:function(sel) {
	var text = this.wikiText();
	text = text.substr(0,sel.selectionStart) + "{START}" + text.substr(sel.selectionStart);
	text = text.substr(0,sel.selectionEnd+7) + "{END}" + text.substr(sel.selectionEnd+7);
	
	//1. Set text with karet holders.
	this.setWikiText(text);
	
	//2. Create a range with the caret holders.
	var range = document.createRange();
	range.setStartAfter($(this.element).find('dks')[0]);
	range.setEndBefore($(this.element).find('dke')[0]);
	
	$(this.element).find('dke').remove();
	$(this.element).find('dks').remove();
	
	var sel = window.getSelection();
	sel.removeAllRanges();
	sel.addRange(range);
	
	// Delete cursor marker
	//document.execCommand('delete', false, null);
},

/*
	Provides ability to insert content into notes.
	This can be used by widgets and other controls to
	insert text/html into the widget.
	
	captureInserted, Boolean indicating if once content is inserted
	selection should include inserted content.
	
	By defualt the caret is always mode to a point after the inserted content.,
	
	if range includes a selection, the selection is replaced with the new contents.
*/	
insert:function(stringOrHtml, captureInserted) {
	var range = this.getSelectedRange();
	var wikitext = this.wikiText();
	if(range) {
		wikitext = wikitext.substr(0, range.location) + stringOrHtml + wikitext.substr(range.location+range.length);
	} else {
		wikitext += stringOrHtml;
	}
	this.setWikiText(wikitext);
	range.location = range.location + $(stringOrHtml).text().length;
	range.length = 0;
	this.setSelectedRange(range);
},

elementByInsertingString:function(stringOrHtml, captureInserted) {
	var range = this.getSelectedRange();
	var wikitext = this.wikiText();
	if(range) {
		wikitext = wikitext.substr(0, range.location) + stringOrHtml + wikitext.substr(range.location+range.length);
	} else {
		wikitext += stringOrHtml;
	}
	return $("<div display='none'>"+this.htmlStringFromOefText(wikitext)+"</div>");
},

paste:function(event) {
	//Enforce plaintext only pasting, in the future we might parse the content
	for(var t in event.originalEvent.clipboardData.types) {
		NSLog("next paste type:"+ event.originalEvent.clipboardData.types[t]);
		if(event.originalEvent.clipboardData.types[t] == 'text/plain') {
			var str = event.originalEvent.clipboardData.getData('text/plain');
			
			//replace all line breaks with <br>
			str = str.replace(/\n/g, '<br>');
			var html = $('<paste>'+str+'</paste>');
			
			//TODO: make universal.
			window.getSelection().getRangeAt(0).insertNode(html[0]);		
		}
	}
	
	event.preventDefault();
},

//Plugin Widgets

addWidget:function(widget) {
	var that = this;
	this.widgets[widget.id] = widget;
	widget.editor = this;
	if(widget.showsButton) {
		$('<a class="dwiki-widget-action" href="#" title="'+widget.tooltip+'">'+widget.char+"</a>")
		.appendTo(this.widgetContainer)
		.mousedown(function(e) {
			//that.applyWidgetAction(widget);
			widget.doActionForEditor(that, that.getSelectedRange());
			e.preventDefault();
			e.stopPropagation();
		}).
		mouseup(function(e) {
			e.preventDefault();
			e.stopPropagation();
		});
	}
},

applyWidgetAction:function(widget, range) {
	if(!range)
		range = this.getSelectedRange();
	var line = 0;//this.getCaretLine();
	var result = widget.apply(range, line, this.wikiText());
		
	this.setWikiText(result[0]);
	this.setSelectedRange(result[1]);
}

}});

/*
	JSContext/Cocoa Mac support
*/
function DWikiBeginMobwriteWithSeed(seed) {
	if(DWiki && DWiki.Editor) {
		DWiki.Editor.seed = seed;
		mobwrite.share(DWiki.Editor);
	}
}

/*	
	-/+ complete incomplete task title
	((comment))
	((@user notified comment))

	h1. Header
	h2. Header two
	h3. Header three
	...
	*Bold*
	_italic_
	`inline code`
	
	[uri Title of link]
	uri
	
	-red-Contents of colored text-/red- where red could also be #htmlhex | rgb(r,g,b)
	#TagName

	Also the order is key for users when writing or generating wiki text. This is highly essetial.
*/

DWiki.HTMLTaskFromWikiLine = function(line) {
	var ret = {};
	ret.changed = 0;
	//- incomplete tasks
	if(line.search(/(^|\n)- (.+)($|\n)/) != -1) { //([\w-\(\)#\*\._ ]+)
		ret.line = line.replace(/(^|\n)- (.+)($|\n)/, "<div class='inline-task'>$2</div>");
		ret.changed = 1;
	}
	//+ complete tasks
	if(line.search(/(^|\n)\+ (.+)($|\n)/) != -1) {
		ret.line = line.replace(/(^|\n)\+ (.+)($|\n)/, "<div class='inline-task complete'>$2</div>");
		ret.changed = 1;
	}
	
	if(!ret.changed) {
		ret.line = line;
	}
	return ret;
}

DWiki.WikiTaskPlainTextContent = function(line) {
	var content = this.HTMLFromWiki(line);
	return $("<div style='display:none'>"+content+"</div>").text();
}

//The task is always refered by its old title
DWiki.WikiTaskUpdateTitle = function(task, title) {
	
}

DWiki.HTMLFromWiki = function(string) {
	//NSLog('converting string:['+string+"]");

	//white space
	//string = string.replace(/ /g, '&nbsp;');

	//First handle all block based syntex items (inline tasks, h1-9)
	var changed = 0;
	var lines = string.split('\n');
	for(var i=0;i<lines.length;i++) {
		var line = lines[i].replace(/\s*$/, '');
		changed = 0;
		
		//tasks
		var res = this.HTMLTaskFromWikiLine(line)
		if(res.changed == 1) {
			line = res.line;
			changed = 1;
		}

		//h1-6 headers
		if(line.search(/(^|\n)h([1-6])\. (.+)($|\n)/) != -1) {
			line = line.replace(/(^|\n)h([1-6])\. (.+)($|\n)/, "<h$2>$3</h$2>");
			changed = 1;
		}
		
		//Block Comments
		if(line.search(/(^|\n)(\(\(([\w\"\'#\_\-\/\*,\.\[\]\: ]+)\)\))($|\n)/) != -1) {
			line = line.replace(/(^|\n)(\(\(([\w\"\'#\_\-\/\*,\.\[\]\: ]+)\)\))($|\n)/, "<div class='comment'><span class='c-text'>$3</span></div>");
			changed = 1;
		}
	
		//Block Targeted comments
		if(line.search(/(^|\n)\(\(@([\w\-]+) ([\w\"\'#\_\-\/\*,\.\[\]\: ]+)\)\)($|\n)/) != -1) {
			line = line.replace(/(^|\n)\(\(@([\w\-]+) ([\w\"\'#\_\-\/\*,\.\[\]\: ]+)\)\)($|\n)/, "<div class='comment' target='$2'><span class='c-target'>$2</span><span class='c-text'>$3</span></div>");
			changed = 1;
		}
		
		//newlines
		if(changed == 0) {
			line = line + "\n<br>";
		}
		lines[i] = line;
	}
	
	string = lines.join('\n');
	
	//Labels
	string = string.replace(/(^| )#([a-zA-Z0-9]+)($|\n| )/g, "<span class='label' id='$2'>$2</span>");
	
	//Comments
	string = string.replace(/\(\(([\w\"\'#\_\-\/\*,\.\[\]\: ]+)\)\)/g, "<span class='comment'><span class='c-text'>$1</span></span>");
	
	//Targeted comments
	string = string.replace(/\(\(@([\w\-]+) ([\w\"\'#\_\-\/\*,\.\[\]\: ]+)\)\)/g, "<span class='comment' target='$1'><span class='c-target'>$1</span><span class='c-text'>$2</span></span>");
	
	//links
	//string = string.replace(/\[(\))
	
	//bold+italic+strike+underline
	string = string.replace(/\*([\w\?\.,\/#_`\-\(\) ]+)\*/g, "<b>$1</b>").replace(/_([\w\?\.,\/#_`\-\(\) ]+)_/g, "<i>$1</i>");

	//Colors
	string = string.replace(/-([\w\#\(\),\.0-9]+)-([\w\?\.,\/# ]+)-\/([\w\#\(\),\.0-9]+)-/g, "<span style='color:$1'>$2</span>");
	
	return string;
}

/*
	Does a best approach to method for converting wiki form plain
*/
DWiki.WikiFromHTML = function(string) {
	string = string.replace(/\<\/dks\>/, "").replace(/\<\/dke\>/, "").replace(/\<dks\>/, "{START}").replace(/\<dke\>/, "{END}");
	string = string.replace(/\<br\>/g, "\n").replace(/&nbsp;/g, " ").replace(/\<paste\>/g, "").replace(/\<\/paste\>/, "");
	return string;
}

//prepares the wiki text for display in a contenteditable container.
DWiki.PrepareWikiText = function(string) {
	string = string.replace(/\n/g, "<br>");
	
	//Caret placements (these get removed on all updates and transmits)
	string = string.replace(/\{START\}/g, "<dks/>").replace(/\{END\}/g, "<dke/>");
	//string = string.replace(/\s/g, "&nbsp;");
	return string;
}

//Returns the wiki text from a content editable
DWiki.WikiTextFromContentEditable = function(element) {
	if($(element).hasClass('entrymode')) {
		if($(element).text() != $(this).attr('placeholder')) {
			return this.StripHtmlFromString(this.WikiFromHTML($(element).html()));
		} else {
			return "";
		}
	} else {
		return this['editable-'+$(element).attr('id')];
	}
}

/*
	Current main setup function to create a DWiki Edit area. This should be depreciated however.
*/
DWiki.WikiFromContentEditable = function(element, wikitext, callback, uid) {
	
	//Create the editor object
	DWiki.Editor = fClass.DWikiEditor.alloc().initWithElement(element, wikitext, uid, callback);
	
	//Legacy setup
//	var that = this;
	
	/*if(wikitext && wikitext.length>0) {
		DWiki.Editor.wikitext = wikitext;	//this['editable-'+$(element).attr('id')] = this.PrepareWikiText(wikitext);
		$(element).empty().html(this.HTMLFromWiki(wikitext));
	} else {
		DWiki.Editor.wikitext = "";//this['editable-'+$(element).attr('id')] = "";
	}*/
	
	//register the update callback
	//this['change-callback-'+$(element).attr('id')] = {callback:callback, ref:ref};

	/*$(element).bind('focus', function(e) {		
		var wikiText = that.PrepareWikiText(DWiki.Editor.wikitext);
		$(this).html(wikiText);
		$(element).addClass('entrymode');
	}).bind('blur', function(e) {
		if($(this).text() != $(this).attr('placeholder')) {	
			var parsedToWikiContent = that.WikiFromHTML($(this).html());
			
			//that['editable-'+$(this).attr('id')] = parsedToWikiContent;
			
			
			$(this).html(that.HTMLFromWiki(parsedToWikiContent));
			
			//Removes wiki editing formattting
			var resultsText = that.StripHtmlFromString(parsedToWikiContent);
			DWiki.Editor.wikitext = resultsText;

			if(callback)
				callback(resultsText, ref);
		} else {
			//that['editable-'+$(this).attr('id')] = "";
			DWiki.Editor.wikitext = "";
			if(callback) {
				callback("", ref);
			}
		}
		
		$(element).removeClass('entrymode');
	}).bind('inline-task-change', function(e) {
		//NSLog('task changed');
		if(callback) {
			callback(DWiki.Editor.wikitext, ref);
		}
	});*/
	
	//Fixes pasting and drag and dropping of notes to force plaintext content
	//this.InforcePlainTextForNoteArea(element);
}

//DEPRECIATED Called remotly
DWiki.RemoteUpdatedEditable = function(element, wikitext) {
	NSLog("Calling RemotelyUpdatedEditable!");
	if(wikitext && wikitext.length>0) {
		if(DWiki.Editor)
			DWiki.Editor.wikitext = wikitext; //this['editable-'+$(element).attr('id')] = wikitext;
		
		$(element).html(this.HTMLFromWiki(wikitext));
	} else {
		if(DWiki.Editor)
			DWiki.Editor.wikitext = ""; //this['editable-'+$(element).attr('id')] = "";
		$(element).empty();
	}
	
}

DWiki.ToggleInlineTask = function(element, title, toggled) {
	var dwiki = DWiki.WikiFromHTML(DWiki.Editor.wikitext /*this['editable-'+$(element).attr('id')])*/);
	//NSLog('getting here with'+dwiki);
	if(dwiki) {
		var lines = dwiki.split('\n');
		for(var i=0;i<lines.length;i++) {
			var line = lines[i];		
			if(line.search(/(^|\n)(-|\+) (.+)($|\n)/) != -1) {
				if(this.WikiTaskPlainTextContent(line) == title) {
					line = line.replace(/(^|\n)(-|\+) (.+)($|\n)/, function(str, $1, $2, $3, $4, $5) {
						//NSLog('string:'+str+':[vs]:'+title);
						//NSLog('checking:['+$1+']['+$2+']['+$3+']['+$4+']['+$5+']');
			
						var nline;
						if(toggled && toggled == true) {
							nline = $1+'+ '+$3+$4;
						} else {
							nline = $1+'- '+$3+$4;
						}
						NSLog('new state:['+nline+']');
						return nline;
					});		
				}
			}
			lines[i] = line;
		}
		//this['editable-'+$(element).attr('id')] = lines.join('\n');
		DWiki.Editor.wikitext = lines.join('\n');
		$(element).trigger('inline-task-change'); //fires the action
	}
}

DWiki.InforcePlainTextForNoteArea = function(element) {
	//inforce plaintext only pasted content.
	
	$(element).bind('paste', function(e) {
		NSLog(e);
	
		for(var t in e.originalEvent.clipboardData.types) {
			if(e.originalEvent.clipboardData.types[t] == 'text/plain') {
				var str = e.originalEvent.clipboardData.getData('text/plain');
				
				//replace all line breaks with <br>
				str = str.replace(/\n/g, '<br>');
				var html = $('<paste>'+str+'</paste>');
				
				window.getSelection().getRangeAt(0).insertNode(html[0]);		
			}
		}
		
		e.preventDefault();
	});
}

//Strips all html from give string.
DWiki.StripHtmlFromString = function(str) {
	return $("<p>"+str+"</p>").text();
}

/*
	Action interface for wiki actions
*/
DWiki.Actions = {}


/***************************************
	Task Title Editing and Actions
****************************************/

//Various widgets that on click do not active complete edit area
$('.inline-task').live('mousedown', function(e) {
	if(e.offsetX < 40) {
		e.stopPropagation();
		e.preventDefault();
		return false;
	}/* else {
		$(this).attr('task', $(this).text());
		e.disableDWikiFocus = true;
		$(this).closest('.dwiki').css('-webkit-user-modify', 'read-only');
		//save the title for editing
	}*/
});

//Handle click states for tasks
$('.inline-task').live('click', function(e) {
	if(e.offsetX < 40) {
		e.stopPropagation();
		e.preventDefault();
		if($(this).hasClass('complete')) {
			$(this).removeClass('complete');
			DWiki.ToggleInlineTask($(this).closest('.dwiki'), $(this).text());
		} else {
			$(this).addClass('complete');
			DWiki.ToggleInlineTask($(this).closest('.dwiki'), $(this).text(), true);
		}
		return false;
	}
});


/***************************************
	Comment Creation and Editing
****************************************/

$('.dwiki .comment').live('mousedown', function(e) {
	e.stopPropagation();
	e.preventDefault();
	return false;
});

/***************************************
				Labels
****************************************/

$('.dwiki .label').live('mousedown', function(e) {
	e.stopPropagation();
	e.preventDefault();
	return false;
});

/*
	Modwrite handler for a DWikiEditor
*/
mobwrite.shareDWikiEditorObj = function(DWikiEditorInst) {	
	// Call our prototype's constructor.
	var key = DWikiEditorInst.syncKey();
	NSLog("Share key:"+key);
	mobwrite.shareObj.apply(this, [key]);
	
	//Set internally the editor
	this.editor = DWikiEditorInst;
};

// The DWikiEditor's shared object's parent is a shareObj.
mobwrite.shareDWikiEditorObj.prototype = new mobwrite.shareObj('');

/**
 * Retrieve the user's wikitext from the editor.
 * This uses the Dwiki function to get the raw inner content of
 * the contenteditable area cleaning up the document as needed.
 * @return {string} Plaintext content.
 */
mobwrite.shareDWikiEditorObj.prototype.getClientText = function() {
	return this.editor.wikiText();
};

/**
 * Set the DWikiEditor's text.
 * @param {string} text New text
 */
mobwrite.shareDWikiEditorObj.prototype.setClientText = function(text) {
	this.editor.setWikiText(text);
	
	/* This is causing an error in -dispatchEvent, cuasing things to crash?? */
	this.fireChange(this.editor.element);
}

mobwrite.shareDWikiEditorObj.prototype.patchClientText = function(patches) {
 // Set some constants which tweak the matching behaviour.
  // Maximum distance to search from expected location.
  this.dmp.Match_Distance = 1000;
  // At what point is no match declared (0.0 = perfection, 1.0 = very loose)
  this.dmp.Match_Threshold = 0.6;

  var oldClientText = this.getClientText();
  
  var cursor = this.captureCursor_();
  // Pack the cursor offsets into an array to be adjusted.
  // See http://neil.fraser.name/writing/cursor/
  var offsets = [];
  if (cursor) {
    offsets[0] = cursor.startOffset;
    if ('endOffset' in cursor) {
      offsets[1] = cursor.endOffset;
    }
  }
  var newClientText = this.patch_apply_(patches, oldClientText, offsets);
  // Set the new text only if there is a change to be made.
  if (oldClientText != newClientText) {
    this.setClientText(newClientText);
    if (cursor) {
      // Unpack the offset array.
      cursor.startOffset = offsets[0];
      if (offsets.length > 1) {
        cursor.endOffset = offsets[1];
        if (cursor.startOffset >= cursor.endOffset) {
          cursor.collapsed = true;
        }
      }
      this.restoreCursor_(cursor);
    }
  }
}

/**
 * Merge a set of patches onto the text.  Return a patched text.
 * @param {Array.<patch_obj>} patches Array of patch objects.
 * @param {string} text Old text.
 * @param {Array.<number>} offsets Offset indices to adjust.
 * @return {string} New text.
 */
mobwrite.shareDWikiEditorObj.prototype.patch_apply_ = function(patches, text, offsets) {
  if (patches.length == 0) {
    return text;
  }

  // Deep copy the patches so that no changes are made to originals.
  patches = this.dmp.patch_deepCopy(patches);
  var nullPadding = this.dmp.patch_addPadding(patches);
  text = nullPadding + text + nullPadding;

  this.dmp.patch_splitMax(patches);
  // delta keeps track of the offset between the expected and actual location
  // of the previous patch.  If there are patches expected at positions 10 and
  // 20, but the first patch was found at 12, delta is 2 and the second patch
  // has an effective expected position of 22.
  var delta = 0;
  for (var x = 0; x < patches.length; x++) {
    var expected_loc = patches[x].start2 + delta;
    var text1 = this.dmp.diff_text1(patches[x].diffs);
    var start_loc;
    var end_loc = -1;
    if (text1.length > this.dmp.Match_MaxBits) {
      // patch_splitMax will only provide an oversized pattern in the case of
      // a monster delete.
      start_loc = this.dmp.match_main(text,
          text1.substring(0, this.dmp.Match_MaxBits), expected_loc);
      if (start_loc != -1) {
        end_loc = this.dmp.match_main(text,
            text1.substring(text1.length - this.dmp.Match_MaxBits),
            expected_loc + text1.length - this.dmp.Match_MaxBits);
        if (end_loc == -1 || start_loc >= end_loc) {
          // Can't find valid trailing context.  Drop this patch.
          start_loc = -1;
        }
      }
    } else {
      start_loc = this.dmp.match_main(text, text1, expected_loc);
    }
    if (start_loc == -1) {
      // No match found.  :(
      if (mobwrite.debug) {
        NSLog('Patch failed: ' + patches[x]);
      }
      // Subtract the delta for this failed patch from subsequent patches.
      delta -= patches[x].length2 - patches[x].length1;
    } else {
      // Found a match.  :)
      if (mobwrite.debug) {
        NSLog('Patch OK.');
      }
      delta = start_loc - expected_loc;
      var text2;
      if (end_loc == -1) {
        text2 = text.substring(start_loc, start_loc + text1.length);
      } else {
        text2 = text.substring(start_loc, end_loc + this.dmp.Match_MaxBits);
      }
      // Run a diff to get a framework of equivalent indices.
      var diffs = this.dmp.diff_main(text1, text2, false);
      if (text1.length > this.dmp.Match_MaxBits &&
          this.dmp.diff_levenshtein(diffs) / text1.length >
          this.dmp.Patch_DeleteThreshold) {
        // The end points match, but the content is unacceptably bad.
        if (mobwrite.debug) {
          window.console.warn('Patch contents mismatch: ' + patches[x]);
        }
      } else {
        var index1 = 0;
        var index2;
        for (var y = 0; y < patches[x].diffs.length; y++) {
          var mod = patches[x].diffs[y];
          if (mod[0] !== DIFF_EQUAL) {
            index2 = this.dmp.diff_xIndex(diffs, index1);
          }
          if (mod[0] === DIFF_INSERT) {  // Insertion
            text = text.substring(0, start_loc + index2) + mod[1] +
                   text.substring(start_loc + index2);
            for (var i = 0; i < offsets.length; i++) {
              if (offsets[i] + nullPadding.length > start_loc + index2) {
                offsets[i] += mod[1].length;
              }
            }
          } else if (mod[0] === DIFF_DELETE) {  // Deletion
            var del_start = start_loc + index2;
            var del_end = start_loc + this.dmp.diff_xIndex(diffs,
                index1 + mod[1].length);
            text = text.substring(0, del_start) + text.substring(del_end);
            for (var i = 0; i < offsets.length; i++) {
              if (offsets[i] + nullPadding.length > del_start) {
                if (offsets[i] + nullPadding.length < del_end) {
                  offsets[i] = del_start - nullPadding.length;
                } else {
                  offsets[i] -= del_end - del_start;
                }
              }
            }
          }
          if (mod[0] !== DIFF_DELETE) {
            index1 += mod[1].length;
          }
        }
      }
    }
  }
  // Strip the padding off.
  text = text.substring(nullPadding.length, text.length - nullPadding.length);
  return text;
};

/**
 * Record information regarding the current cursor.
 * @return {Object?} Context information of the cursor.
 * @private
 */
mobwrite.shareDWikiEditorObj.prototype.captureCursor_ = function() {
  if ('activeElement' in this.editor.element && !this.editor.element.activeElement) {
    // Safari specific code.
    // Restoring a cursor in an unfocused element causes the focus to jump.
    return null;
  }
  var padLength = this.dmp.Match_MaxBits / 2;  // Normally 16.
  var text = this.editor.wikiText();
  var cursor = {};
  var sel = this.editor.getSelectionOffsets();
  if (this.editor.getRange) {  // W3
    try {
      var selectionStart = sel.selectionStart;
      var selectionEnd = sel.selectionEnd;
    } catch (e) {
      // No cursor; the element may be "display:none".
      return null;
    }
    cursor.startPrefix = text.substring(selectionStart - padLength, selectionStart);
    cursor.startSuffix = text.substring(selectionStart, selectionStart + padLength);
    cursor.startOffset = selectionStart;
    cursor.collapsed = (selectionStart == selectionEnd);
    if (!cursor.collapsed) {
      cursor.endPrefix = text.substring(selectionEnd - padLength, selectionEnd);
      cursor.endSuffix = text.substring(selectionEnd, selectionEnd + padLength);
      cursor.endOffset = selectionEnd;
    }
  } else {  // IE TODO: Not currently implemented
    // Walk up the tree looking for this textarea's document node.
    var doc = this.element;
    while (doc.parentNode) {
      doc = doc.parentNode;
    }
    if (!doc.selection || !doc.selection.createRange) {
      // Not IE?
      return null;
    }
    var range = doc.selection.createRange();
    if (range.parentElement() != this.element) {
      // Cursor not in this textarea.
      return null;
    }
    var newRange = doc.body.createTextRange();

    cursor.collapsed = (range.text == '');
    newRange.moveToElementText(this.element);
    if (!cursor.collapsed) {
      newRange.setEndPoint('EndToEnd', range);
      cursor.endPrefix = newRange.text;
      cursor.endOffset = cursor.endPrefix.length;
      cursor.endPrefix = cursor.endPrefix.substring(cursor.endPrefix.length - padLength);
    }
    newRange.setEndPoint('EndToStart', range);
    cursor.startPrefix = newRange.text;
    cursor.startOffset = cursor.startPrefix.length;
    cursor.startPrefix = cursor.startPrefix.substring(cursor.startPrefix.length - padLength);

    newRange.moveToElementText(this.element);
    newRange.setEndPoint('StartToStart', range);
    cursor.startSuffix = newRange.text.substring(0, padLength);
    if (!cursor.collapsed) {
      newRange.setEndPoint('StartToEnd', range);
      cursor.endSuffix = newRange.text.substring(0, padLength);
    }
  }

  // Record scrollbar locations
  if ('scrollTop' in this.editor.element) {
    cursor.scrollTop = this.editor.element.scrollTop / this.editor.element.scrollHeight;
    cursor.scrollLeft = this.editor.element.scrollLeft / this.editor.element.scrollWidth;
  }

  // alert(cursor.startPrefix + '|' + cursor.startSuffix + ' ' +
  //     cursor.startOffset + '\n' + cursor.endPrefix + '|' +
  //     cursor.endSuffix + ' ' + cursor.endOffset + '\n' +
  //     cursor.scrollTop + ' x ' + cursor.scrollLeft);
  return cursor;
};

/**
 * Attempt to restore the cursor's location.
 * @param {Object} cursor Context information of the cursor.
 * @private
 */
mobwrite.shareDWikiEditorObj.prototype.restoreCursor_ = function(cursor) {
  
  // Set some constants which tweak the matching behaviour.
  // Maximum distance to search from expected location.
  this.dmp.Match_Distance = 1000;
  // At what point is no match declared (0.0 = perfection, 1.0 = very loose)
  this.dmp.Match_Threshold = 0.9;

  var padLength = this.dmp.Match_MaxBits / 2;  // Normally 16.
  var newText = this.editor.wikiText();

  // Find the start of the selection in the new text.
  var pattern1 = cursor.startPrefix + cursor.startSuffix;
  var pattern2, diff;
  var cursorStartPoint = this.dmp.match_main(newText, pattern1,
      cursor.startOffset - padLength);
  if (cursorStartPoint !== null) {
    pattern2 = newText.substring(cursorStartPoint,
                                 cursorStartPoint + pattern1.length);
    //alert(pattern1 + '\nvs\n' + pattern2);
    // Run a diff to get a framework of equivalent indicies.
    diff = this.dmp.diff_main(pattern1, pattern2, false);
    cursorStartPoint += this.dmp.diff_xIndex(diff, cursor.startPrefix.length);
  }

  var cursorEndPoint = null;
  if (!cursor.collapsed) {
    // Find the end of the selection in the new text.
    pattern1 = cursor.endPrefix + cursor.endSuffix;
    cursorEndPoint = this.dmp.match_main(newText, pattern1,
        cursor.endOffset - padLength);
    if (cursorEndPoint !== null) {
      pattern2 = newText.substring(cursorEndPoint,
                                   cursorEndPoint + pattern1.length);
      //alert(pattern1 + '\nvs\n' + pattern2);
      // Run a diff to get a framework of equivalent indicies.
      diff = this.dmp.diff_main(pattern1, pattern2, false);
      cursorEndPoint += this.dmp.diff_xIndex(diff, cursor.endPrefix.length);
    }
  }

  // Deal with loose ends
  if (cursorStartPoint === null && cursorEndPoint !== null) {
    // Lost the start point of the selection, but we have the end point.
    // Collapse to end point.
    cursorStartPoint = cursorEndPoint;
  } else if (cursorStartPoint === null && cursorEndPoint === null) {
    // Lost both start and end points.
    // Jump to the offset of start.
    cursorStartPoint = cursor.startOffset;
  }
  if (cursorEndPoint === null) {
    // End not known, collapse to start.
    cursorEndPoint = cursorStartPoint;
  }

  // Restore selection.
  if (this.editor.getRange) {  // W3
  //  this.element.selectionStart = cursorStartPoint;
  //  this.element.selectionEnd = cursorEndPoint;
  	var sel = {};
  	sel.selectionStart = cursorStartPoint;
  	sel.selectionEnd = cursorEndPoint;
  	this.editor.setSelectionOffsets(sel);
  
  } else {  // IE
    // Walk up the tree looking for this textarea's document node.
    var doc = this.element;
    while (doc.parentNode) {
      doc = doc.parentNode;
    }
    if (!doc.selection || !doc.selection.createRange) {
      // Not IE?
      return;
    }
    // IE's TextRange.move functions treat '\r\n' as one character.
    var snippet = this.element.value.substring(0, cursorStartPoint);
    var ieStartPoint = snippet.replace(/\r\n/g, '\n').length;

    var newRange = doc.body.createTextRange();
    newRange.moveToElementText(this.element);
    newRange.collapse(true);
    newRange.moveStart('character', ieStartPoint);
    if (!cursor.collapsed) {
      snippet = this.element.value.substring(cursorStartPoint, cursorEndPoint);
      var ieMidLength = snippet.replace(/\r\n/g, '\n').length;
      newRange.moveEnd('character', ieMidLength);
    }
    newRange.select();
  }

  // Restore scrollbar locations
  if ('scrollTop' in cursor) {
    this.editor.element.scrollTop = cursor.scrollTop * this.editor.element.scrollHeight;
    this.editor.element.scrollLeft = cursor.scrollLeft * this.editor.element.scrollWidth;
  }
};

/*//Called prior to doing a local patch from a patchs above, and carets have been encoded.
mobwrite.shareDWikiEditorObj.prototype.saveCarets_ = function() {
	var closRange =  this.dnote.editor.getRange();
	if(closRange) {
		//Create the carets, google inserts <span id='goog_...'> for the carets.
		this._savedCarets = new goog.editor.range.NormalizedCaretRange_(closRange);
	}
};

//Called after doing a local patch from a patchs above, and carets have been decoded, and new text applied to doc.
mobwrite.shareDWikiEditorObj.prototype.restoreCarets_ = function() {
	if(this._savedCarets) {
		this._savedCarets.setRestorationDocument(this.dnote.editor.getEditableDomHelper().getDocument());
		this._savedCarets.restore();
		this._savedCarets = null;
	}
};

//Removes carets without resoring them
mobwrite.shareDWikiEditorObj.prototype.removeCarets_ = function() {
	this._savedCarets.setRestorationDocument(this.dnote.editor.getEditableDomHelper().getDocument());
	this._savedCarets.remove();
	this._savedCarets = null;
}*/

/**
 * Handler to accept editor instances
*/
mobwrite.shareDWikiEditorObj.shareHandler = function(editor) {
	return new mobwrite.shareDWikiEditorObj(editor);
}

// Register this shareHandler with MobWrite.
mobwrite.shareHandlers.push(mobwrite.shareDWikiEditorObj.shareHandler);